# Copyright (c) Microsoft. All rights reserved.
# Modified the code for ResNet training in the BLBF setting. Thorsten Joachims, October 2017

# Licensed under the MIT license. See LICENSE.md file in the project root
# for full license information.
# ==============================================================================

from __future__ import print_function
import os
import argparse
import sys

import numpy as np
import cntk as C
from cntk import cross_entropy_with_softmax, classification_error, reduce_mean, softmax, times_transpose, negate, dropout, element_divide, element_times, plus, reduce_mean
from cntk.io import MinibatchSource, ImageDeserializer, CTFDeserializer, StreamDef, StreamDefs
import cntk.io.transforms as xforms
from cntk import Trainer, Evaluator, cntk_py
from cntk.learners import momentum_sgd, adam, learning_rate_schedule, momentum_as_time_constant_schedule, UnitType
from cntk.debugging import set_computation_network_trace_level
from cntk.logging import *
from cntk.debugging import *
from resnet_models import *
from cntk.ops.functions import load_model

# Paths relative to current python file.
abs_path   = os.path.dirname(os.path.abspath(__file__))
data_path  = os.path.join(abs_path, "CIFAR-10")

# model dimensions
image_height = 32
image_width  = 32
num_channels = 3  # RGB
num_classes  = 10

# Define the reader for both training and evaluation action.
def create_train_reader(img_file, blbf_file, mean_file):
    if not os.path.exists(img_file) or not os.path.exists(mean_file) or not os.path.exists(blbf_file):
        raise RuntimeError("File '%s' or '%s' or '%s' does not exist. Please run install_cifar10.py from DataSets/CIFAR-10 to fetch them" %
                           (img_file, blbf_file, mean_file))

    # transformation pipeline for the features has jitter/crop only when training
    transforms = []
    transforms += [
        xforms.crop(crop_type='randomside', side_ratio=0.8, jitter_type='uniratio') # train uses jitter
    ]
    transforms += [
        xforms.scale(width=image_width, height=image_height, channels=num_channels, interpolations='linear'),
        xforms.mean(mean_file)
    ]
    # deserializer
    image_source=ImageDeserializer(img_file, StreamDefs(
        features=StreamDef(field='image', transforms=transforms), 
        labels=StreamDef(field='label', shape=num_classes)))
    blbf_source=CTFDeserializer(blbf_file, StreamDefs(
#        ids=StreamDef(field='id', shape=1),
#        actions=StreamDef(field='action', shape=1),
#        fulllabels=StreamDef(field='fulllabel', shape=1),
#        probs=StreamDef(field='probs', shape=num_classes),
        actionvecs=StreamDef(field='actionvec', shape=num_classes, is_sparse=True),
        losses=StreamDef(field='loss', shape=1),
        props=StreamDef(field='prop', shape=1)))
    return MinibatchSource([image_source,blbf_source],randomize=True)


def create_val_reader(img_file, blbf_file, mean_file):
    if not os.path.exists(img_file) or not os.path.exists(mean_file) or not os.path.exists(blbf_file):
        raise RuntimeError("File '%s' or '%s' or '%s' does not exist. Please run install_cifar10.py from DataSets/CIFAR-10 to fetch them" %
                           (img_file, blbf_file, mean_file))

    # transformation pipeline for the features has jitter/crop only when training
    transforms = []
    transforms += [
        xforms.scale(width=image_width, height=image_height, channels=num_channels, interpolations='linear'),
        xforms.mean(mean_file)
    ]
    # deserializer
    image_source=ImageDeserializer(img_file, StreamDefs(
        features=StreamDef(field='image', transforms=transforms), 
        labels=StreamDef(field='label', shape=num_classes)))
    blbf_source=CTFDeserializer(blbf_file, StreamDefs(
#        ids=StreamDef(field='id', shape=1),
#        actions=StreamDef(field='action', shape=1),
#        fulllabels=StreamDef(field='fulllabel', shape=1),
#        probs=StreamDef(field='probs', shape=num_classes),
        actionvecs=StreamDef(field='actionvec', shape=num_classes, is_sparse=True),
        losses=StreamDef(field='loss', shape=1),
        props=StreamDef(field='prop', shape=1)))
    return MinibatchSource([image_source,blbf_source],randomize=True)


def create_test_reader(map_file, mean_file):
    if not os.path.exists(map_file) or not os.path.exists(mean_file):
        raise RuntimeError("File '%s' or '%s' does not exist. Please run install_cifar10.py from DataSets/CIFAR-10 to fetch them" %
                           (map_file, mean_file))

    # transformation pipeline for the features has jitter/crop only when training
    transforms = []
    transforms += [
        xforms.scale(width=image_width, height=image_height, channels=num_channels, interpolations='linear'),
        xforms.mean(mean_file)
    ]
    # deserializer
    return MinibatchSource(ImageDeserializer(map_file, StreamDefs(
        features=StreamDef(field='image', transforms=transforms), # first column in map file is referred to as 'image'
        labels=StreamDef(field='label', shape=num_classes))))   # and second as 'label'


# Train and evaluate the network.
def train_and_evaluate(reader_train, reader_trainval, reader_val, reader_test, network_name, epoch_size, max_epochs, loss_fn, l2_reg_weight=0.0001, minibatch_size=128, learn_rate=0.1, lagrange_mult=0.0, profiler_dir=None,
                       model_dir=None, tensorboard_logdir=None, start_checkpoint=None):

    set_computation_network_trace_level(0)

    # Input variables denoting the features and label data
    input_var = C.input_variable((num_channels, image_height, image_width))
    label_var = C.input_variable((num_classes))
#    id_var = C.input_variable((1))
#    action_var = C.input_variable((1))
#    fulllabel_var = C.input_variable((1))
#    probs_var = C.input_variable((num_classes))
    actionvec_var = C.input_variable((num_classes))
    loss_var = C.input_variable((1))
    prop_var = C.input_variable((1))


    # create model, and configure learning parameters
    if network_name == 'resnet20':
        z = create_cifar10_model(input_var, 3, num_classes)
        lr_per_mb = [learn_rate*1.0]*300+[learn_rate*0.1]*200+[learn_rate*0.01]
    elif network_name == 'resnet110':
        z = create_cifar10_model(input_var, 18, num_classes)
        lr_per_mb = [learn_rate*0.1]*1+[learn_rate*1.0]*300+[learn_rate*0.1]*200+[learn_rate*0.01]
    else:
        return RuntimeError("Unknown model name!")

    # loss and metric
    ce = lagrange_mult+element_divide(element_times(times_transpose(softmax(z),actionvec_var),loss_var-lagrange_mult),prop_var) # erm risk with loss offset
    ips = plus(element_divide(element_times(times_transpose(softmax(z),actionvec_var),loss_var),prop_var),reduce_mean(label_var)*0.0) # ips risk
    pe = classification_error(z, label_var) # hard classification error of argmax prediction
    petrain = plus(plus(plus(classification_error(z, label_var),reduce_mean(actionvec_var)*0.0),loss_var*0.0),prop_var*0.0) # hard classification error
    spetrain = 1 + plus(plus(plus(negate(times_transpose(softmax(z), label_var)),reduce_mean(actionvec_var)*0.0),loss_var*0.0),prop_var*0.0) # probabilistic rule error
    spetest = 1 + negate(times_transpose(softmax(z), label_var)) # probabilistic rule error
    sn = plus(plus(element_divide(times_transpose(softmax(z),actionvec_var),prop_var),loss_var*0.0),reduce_mean(label_var)*0.0) # denominator of SNIPS
    ips_sq = C.pow(plus(element_divide(element_times(times_transpose(softmax(z),actionvec_var),loss_var),prop_var),reduce_mean(label_var)*0.0),2.0) # square of ips risk
    sn_sq = C.pow(plus(plus(element_divide(times_transpose(softmax(z),actionvec_var),prop_var),loss_var*0.0),reduce_mean(label_var)*0.0),2.0) # square of denominator of SNIPS

    
    # shared training parameters
#    minibatch_size = 128
    momentum_time_constant = -minibatch_size/np.log(0.9)
#    l2_reg_weight = 0.00001

    # Set learning parameters
    lr_per_sample = [lr/minibatch_size for lr in lr_per_mb]
    lr_schedule = learning_rate_schedule(lr_per_sample, epoch_size=epoch_size, unit=UnitType.sample)
    mm_schedule = momentum_as_time_constant_schedule(momentum_time_constant)

    # progress writers
    progress_writers = [ProgressPrinter(tag='Training', num_epochs=max_epochs)]
    progress_writers_ips = [ProgressPrinter(tag='IPS', num_epochs=max_epochs)]
    progress_writers_sn = [ProgressPrinter(tag='SelfNormalizer', num_epochs=max_epochs)]
    progress_writers_ipsval = [ProgressPrinter(tag='IPSVal', num_epochs=max_epochs)]
    progress_writers_snval = [ProgressPrinter(tag='SelfNormalizerVal', num_epochs=max_epochs)]
    progress_writers_ipssqval = [ProgressPrinter(tag='IPSVal_SQ', num_epochs=max_epochs)]
    progress_writers_snsqval = [ProgressPrinter(tag='SelfNormalizerVal_SQ', num_epochs=max_epochs)]
    progress_writers_petrain = [ProgressPrinter(tag='ErrorTrain', num_epochs=max_epochs)]
    progress_writers_peval = [ProgressPrinter(tag='ErrorVal', num_epochs=max_epochs)]
    progress_writers_petest = [ProgressPrinter(tag='ErrorTest', num_epochs=max_epochs)]
    progress_writers_spetrain = [ProgressPrinter(tag='StochasticErrorTrain', num_epochs=max_epochs)]
    progress_writers_speval = [ProgressPrinter(tag='StochasticErrorVal', num_epochs=max_epochs)]
    progress_writers_spetest = [ProgressPrinter(tag='StochasticErrorTest', num_epochs=max_epochs)]
    tensorboard_writer = None
    if tensorboard_logdir is not None:
        tensorboard_writer = TensorBoardProgressWriter(freq=10, log_dir=tensorboard_logdir, model=z)
        progress_writers.append(tensorboard_writer)

    # trainer object
    learner = momentum_sgd(z.parameters, lr_schedule, mm_schedule,
                           l2_regularization_weight = l2_reg_weight)                           
    trainer = Trainer(z, (ce, pe), learner, progress_writers)
    ips_evaluator = Evaluator(ips, progress_writers_ips)
    sn_evaluator = Evaluator(sn, progress_writers_sn)
    ipsval_evaluator = Evaluator(ips, progress_writers_ipsval)
    snval_evaluator = Evaluator(sn, progress_writers_snval)
    ipssqval_evaluator = Evaluator(ips_sq, progress_writers_ipssqval)
    snsqval_evaluator = Evaluator(sn_sq, progress_writers_snsqval)
    petrain_evaluator = Evaluator(petrain, progress_writers_petrain)
    peval_evaluator = Evaluator(petrain, progress_writers_peval)
    petest_evaluator = Evaluator(pe, progress_writers_petest)
    spetrain_evaluator = Evaluator(spetrain, progress_writers_spetrain)
    speval_evaluator = Evaluator(spetrain, progress_writers_speval)
    spetest_evaluator = Evaluator(spetest, progress_writers_spetest)
    
    # define mapping from reader streams to network inputs
    input_map_train = {
        input_var: reader_train.streams.features,
        label_var: reader_train.streams.labels,
#        id_var: reader_train.streams.ids,
#        action_var: reader_train.streams.actions,
#        fulllabel_var: reader_train.streams.fulllabels,
#        probs_var: reader_train.streams.probs,
        actionvec_var: reader_train.streams.actionvecs,
        loss_var: reader_train.streams.losses,
        prop_var: reader_train.streams.props
    }
    input_map_test = {
        input_var: reader_train.streams.features,
        label_var: reader_train.streams.labels,
    }

    log_number_of_parameters(z) ; print()

    # perform model training
    if profiler_dir:
        start_profiler(profiler_dir, True)

    if start_checkpoint and os.path.exists(start_checkpoint):
        trainer.restore_from_checkpoint(start_checkpoint)
        
    for epoch in range(max_epochs):       # loop over epochs
        sample_count = 0
        while sample_count < epoch_size:  # loop over minibatches in the epoch
            data = reader_train.next_minibatch(min(minibatch_size, epoch_size-sample_count), input_map=input_map_train) # fetch minibatch.
#            ips_evaluator.test_minibatch(data)
#            sn_evaluator.test_minibatch(data)
#            petrain_evaluator.test_minibatch(data)
#            spetrain_evaluator.test_minibatch(data)
            trainer.train_minibatch(data)                                   # update model with it
            sample_count += trainer.previous_minibatch_sample_count         # count samples processed so far

        trainer.summarize_training_progress()
#        ips_evaluator.summarize_test_progress()
#        sn_evaluator.summarize_test_progress()
#        petrain_evaluator.summarize_test_progress()
#        spetrain_evaluator.summarize_test_progress()

        # Log mean of each parameter tensor, so that we can confirm that the parameters change indeed.
        if tensorboard_writer:
            for parameter in z.parameters:
                tensorboard_writer.write_value(parameter.uid + "/mean", reduce_mean(parameter).eval(), epoch)

        if model_dir:
            trainer.save_checkpoint(os.path.join(model_dir, network_name + "_" + loss_fn + "_" + str(l2_reg_weight) + "_" + str(minibatch_size) + "_" + str(learn_rate) + "_" + str(lagrange_mult) + "_{}".format(epoch)))
        enable_profiler() # begin to collect profiler data after first epoch
        
        if (epoch + 1) % 10 == 0:               # Evaluate the current model on the training set
            # Evaluation parameters
            test_epoch_size     = epoch_size
            sample_count    = 0

            while sample_count < test_epoch_size:
                current_minibatch = min(minibatch_size, test_epoch_size - sample_count)
                # Fetch next train mini batch from the second train reader.
                data = reader_trainval.next_minibatch(current_minibatch, input_map=input_map_train)
                # Keep track of the number of samples processed so far.
                sample_count += data[label_var].num_samples
                ips_evaluator.test_minibatch(data)
                sn_evaluator.test_minibatch(data)
                petrain_evaluator.test_minibatch(data)
                spetrain_evaluator.test_minibatch(data)
                
            print("")    
            print("Training Set (IPS Estimate of Error)")
            ips_evaluator.summarize_test_progress()
            print("Training Set (Self-Normalizer)")
            sn_evaluator.summarize_test_progress()
            print("Training Set (Full-Information Error of Argmax Policy)")
            petrain_evaluator.summarize_test_progress()
            print("Training Set (Full-Information Error of Stochastic Policy)")
            spetrain_evaluator.summarize_test_progress()
            print("")
            
        sys.stdout.flush()


        if (epoch + 1) % 10 == 0 and reader_val:             # Evaluate the current model on the validation set
            # Evaluation parameters
            test_epoch_size     = 15000 # WARNING: need to fill in the right size of the validation set here.
            sample_count    = 0

            while sample_count < test_epoch_size:
                current_minibatch = min(minibatch_size, test_epoch_size - sample_count)
                # Fetch next validation mini batch.
                data = reader_val.next_minibatch(current_minibatch, input_map=input_map_train)
                # Keep track of the number of samples processed so far.
                sample_count += data[label_var].num_samples
                ipsval_evaluator.test_minibatch(data)
                snval_evaluator.test_minibatch(data)
                ipssqval_evaluator.test_minibatch(data)
                snsqval_evaluator.test_minibatch(data)
                peval_evaluator.test_minibatch(data)
                speval_evaluator.test_minibatch(data)
                
            print("Validation Set (IPS Estimate of Error)")
            ipsval_evaluator.summarize_test_progress()
            print("Validation Set (Self-Normalizer)")
            snval_evaluator.summarize_test_progress()
            print("Validation Set (IPS Estimate of Error Squared)")
            ipssqval_evaluator.summarize_test_progress()
            print("Validation Set (Self-Normalizer Squared)")
            snsqval_evaluator.summarize_test_progress()
            print("Validation Set (Full-Information Error of Argmax Policy)")
            peval_evaluator.summarize_test_progress()
            print("Validation Set (Full-Information Error of Stochastic Policy)")
            speval_evaluator.summarize_test_progress()
            print("")
            
        sys.stdout.flush()

        if (epoch + 1) % 10 == 0:                 # Evaluate the current model on the test set.
            # Evaluation parameters
            test_epoch_size     = 10000 # WARNING: need to fill in the right size of the test set here.

            # process minibatches and evaluate the model
            metric_numer    = 0
            metric_denom    = 0
            sample_count    = 0

            while sample_count < test_epoch_size:
                current_minibatch = min(minibatch_size, test_epoch_size - sample_count)
                # Fetch next test min batch.
                data = reader_test.next_minibatch(current_minibatch, input_map=input_map_test)
                # minibatch data to be trained with
                metric_numer += trainer.test_minibatch(data) * current_minibatch
                metric_denom += current_minibatch
                # Keep track of the number of samples processed so far.
                sample_count += data[label_var].num_samples
                spetest_evaluator.test_minibatch(data)

            print("Test Set (Full-Information Error of Argmax Policy)")
            trainer.summarize_test_progress()
            print("Test Set (Full-Information Error of Stochastic Policy)")
            spetest_evaluator.summarize_test_progress()
            print(metric_numer/metric_denom)
            print("")
            
        sys.stdout.flush()

    if profiler_dir:
        stop_profiler()

    if model_dir:
        z.save(os.path.join(model_dir, "model_" + network_name + "_" + loss_fn + "_" + str(l2_reg_weight) + "_" + str(minibatch_size) + "_" + str(learn_rate) + "_" + str(lagrange_mult) + "_{}".format(epoch)))

    return 1


if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-x', '--train_img_file', help='Full-information training file with images and full information label for debugging.', required=True, default="./CIFAR-10-BLBF/train_map_full1.txt")
    parser.add_argument('-y', '--train_blbf_file', help='BLBF training file with losses and propensities.', required=True, default="./CIFAR-10-BLBF/train_map_blbf1.txt")
    parser.add_argument('-v', '--val_img_file', help='Full-information validation file with images.', required=False, default=None)
    parser.add_argument('-w', '--val_blbf_file', help='BLBF validation file with losses and propensities.', required=False, default=None)
    parser.add_argument('-z', '--test_file', help='Full-information test file.', required=True, default="./CIFAR-10/test_map.txt")

    parser.add_argument('-n', '--network', help='Network type (resnet20 or resnet110)', required=False, default='resnet20')
    parser.add_argument('-c', '--l2_reg_weight', help='L2 regularization parameter', required=False, default='0.0001')
    parser.add_argument('-b', '--minibatch_size', help='Minibatch size', required=False, default='128')
    parser.add_argument('-r', '--learn_rate', help='Factor for learning rate', required=False, default='0.1')
    parser.add_argument('-l', '--lagrange_mult', help='Lagrange multiplier value to subtract from the loss', required=False, default='0.0')
    parser.add_argument('-e', '--epochs', help='Number of training epochs', required=False, default='1000')
    parser.add_argument('-p', '--profiler_dir', help='Directory for saving profiler output', required=False, default=None)
    parser.add_argument('-m', '--model_dir', help='Directory for saving model', required=False, default=None)
    parser.add_argument('-tensorboard_logdir', '--tensorboard_logdir', help='Directory where TensorBoard logs should be created', required=False, default=None)
    parser.add_argument('-s', '--start_checkpoint', help='Checkpointed model to start training with.', required=False, default=None)

    args = vars(parser.parse_args())
    epochs = int(args['epochs'])
    network_name = args['network']
    l2_reg_weight = float(args['l2_reg_weight'])
    minibatch_size = int(args['minibatch_size'])
    learn_rate = float(args['learn_rate'])
    lagrange_mult = float(args['lagrange_mult'])
    loss_fn="blbf"
    
#    model_dir = args['model_dir']
#    if not model_dir:
#        model_dir = os.path.join(abs_path, "Models")

    reader_train = create_train_reader(args['train_img_file'], args['train_blbf_file'], os.path.join(data_path, 'CIFAR-10_mean.xml')) # create reader for training set
    reader_trainval = create_val_reader(args['train_img_file'], args['train_blbf_file'], os.path.join(data_path, 'CIFAR-10_mean.xml')) # create a second reader for the training set to make it easier to compute training statistics
    if args['val_img_file'] and args['val_blbf_file']:
        reader_val = create_val_reader(args['val_img_file'], args['val_blbf_file'], os.path.join(data_path, 'CIFAR-10_mean.xml')) # create reader for computing validation set
    else:
        reader_val = False
    reader_test  = create_test_reader(args['test_file'], os.path.join(data_path, 'CIFAR-10_mean.xml')) # create reader for full-information test set

    with open(args['train_img_file'],"r") as f:
        lines=f.readlines()
    epoch_size = len(lines)
    print("Epoch Size = "+str(epoch_size))
    print(args)
    sys.stdout.flush()
    
    train_and_evaluate(reader_train, reader_trainval, reader_val, reader_test, network_name, epoch_size, epochs, loss_fn, l2_reg_weight, minibatch_size, learn_rate, lagrange_mult, args['profiler_dir'], args['model_dir'],args['tensorboard_logdir'],args['start_checkpoint'])

